use anyhow::Result;
use std::convert::TryFrom;
use thiserror::Error;

static INPUT: &str = include_str!("../../input08");

fn main() -> Result<()> {
    let mut handheld = Handheld::try_from(INPUT)?;
    let acc = handheld.run().unwrap_looped();
    println!("acc: {}", acc);

    let mut handheld = Handheld::try_from(INPUT)?;
    let acc = handheld.run_fixed();
    println!("acc: {}", acc);
    Ok(())
}

#[derive(Error, Debug)]
enum Errors<'a> {
    #[error("invalid input ({0}): \"{1}\"")]
    InvalidInput(&'a str, &'a str),
}

#[derive(Debug, Clone)]
struct Handheld {
    pos: usize,
    game: Vec<Op>,
    visited: Vec<bool>,
    acc: usize,
}

impl<'a> TryFrom<&'a str> for Handheld {
    type Error = Errors<'a>;

    fn try_from(game: &'a str) -> Result<Self, Self::Error> {
        let game = game
            .lines()
            .map(|line| Op::try_from(line))
            .collect::<Result<Vec<Op>, Errors>>()?;
        let visited = vec![false; game.len()];
        Ok(Handheld {
            pos: 0,
            game,
            visited,
            acc: 0,
        })
    }
}

impl Handheld {
    fn run_fixed(&mut self) -> usize {
        loop {
            let pos = self.pos;
            match self.visited.get(pos) {
                Some(false) => (),
                Some(true) => panic!("looped"),
                None => panic!("out of range"),
            };
            if let Some(acc) = self.step_fixed() {
                return acc;
            }
        }
    }

    fn step_fixed(&mut self) -> Option<usize> {
        let pos = self.pos;
        let op = self.game[pos];

        let patched = match op {
            Op::Acc(_) => None,
            Op::Jump(delta) => {
                let mut patched = self.clone();
                patched.game[pos] = Op::Nop(delta);
                Some(patched)
            }
            Op::Nop(delta) => {
                let mut patched = self.clone();
                patched.game[pos] = Op::Jump(delta);
                Some(patched)
            }
        };
        if let Some(mut patched) = patched {
            if let HandheldStatus::Finished(acc) = patched.run() {
                return Some(acc);
            }
        }

        self.step();
        None
    }

    fn run(&mut self) -> HandheldStatus {
        loop {
            let pos = self.pos;
            if pos == self.game.len() {
                return HandheldStatus::Finished(self.acc);
            }
            match self.visited.get(pos) {
                Some(false) => (),
                Some(true) => return HandheldStatus::Looped(self.acc),
                None => return HandheldStatus::Failed,
            };
            self.step();
        }
    }

    fn step(&mut self) {
        let pos = self.pos;
        let op = self.game[pos];
        match op {
            Op::Acc(delta) => {
                self.acc = (self.acc as i32 + delta) as usize;
                self.pos += 1
            }
            Op::Jump(delta) => self.pos = (self.pos as i32 + delta) as usize,
            Op::Nop(_) => self.pos += 1,
        }
        self.visited[pos] = true;
    }
}

#[derive(Debug, Copy, Clone)]
enum Op {
    Acc(i32),
    Jump(i32),
    Nop(i32),
}

impl<'a> TryFrom<&'a str> for Op {
    type Error = Errors<'a>;

    fn try_from(line: &'a str) -> Result<Self, Self::Error> {
        let (op, delta) = line.split_at(4);
        let delta = delta
            .parse()
            .or(Err(Errors::InvalidInput("invalid number", delta)))?;
        let op = match op {
            "acc " => Op::Acc(delta),
            "jmp " => Op::Jump(delta),
            "nop " => Op::Nop(delta),
            _ => return Err(Errors::InvalidInput("invalid opcode", &op)),
        };
        Ok(op)
    }
}

#[derive(Debug)]
enum HandheldStatus {
    Looped(usize),
    Finished(usize),
    Failed,
}

impl HandheldStatus {
    fn unwrap_looped(self) -> usize {
        match self {
            HandheldStatus::Looped(value) => return value,
            _ => panic!("handhelt status is Finished not Looped"),
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn example() {
        let input = "
nop +0
acc +1
jmp +4
acc +3
jmp -3
acc -99
acc +1
jmp -4
acc +6
"
        .trim();
        let mut handheld = Handheld::try_from(input).unwrap();
        let acc = handheld.run();
        assert_eq!(acc.unwrap_looped(), 5);
    }

    #[test]
    fn example2() {
        let input = "
nop +0
acc +1
jmp +4
acc +3
jmp -3
acc -99
acc +1
jmp -4
acc +6
"
        .trim();
        let mut handheld = Handheld::try_from(input).unwrap();
        let acc = handheld.run_fixed();
        assert_eq!(acc, 8);
    }

    #[test]
    fn failures() {
        let mut handheld = Handheld::try_from(INPUT).unwrap();
        let acc = handheld.run();
        assert!(acc.unwrap_looped() < 1748);
    }

    #[test]
    fn part1() {
        let mut handheld = Handheld::try_from(INPUT).unwrap();
        let acc = handheld.run();
        assert_eq!(acc.unwrap_looped(), 1709);
    }
}
